#include <string.h>
#ifdef sun
#include <alloca.h>
#endif
#ifdef vax
#include <sys/types.h>
#endif
#include <sys/stat.h>
#include "config.h"
#include "lint.h"
#include "array.h"
#include "interpret.h"
#include "object.h"
#include "wiz_list.h"
#include "regexp.h"
#include "exec.h"
#include "main.h"
#include "stralloc.h"


#define VECTOR_SIZE(nelem) (sizeof(struct vector) +sizeof(struct svalue) * (nelem-1))
#define ALLOC_VECTOR(nelem)  (struct vector *)xalloc(VECTOR_SIZE(nelem))

/*
 * This file contains functions used to manipulate arrays.
 * Some of them are connected to efuns, and some are only used internally
 * by the game driver.
 */

extern void remove_array_from_stack(struct vector *);
extern int d_flag;
extern struct svalue const0;

int num_arrays;
int total_array_size;

/*
 * Make an empty vector for everyone to use, never to be deallocated.
 * It is cheaper to reuse it, than to use malloc() and allocate.
 */
struct vector null_vector = {
    0,	/* size */
    0,  /* malloced size */
    1,	/* Ref count, which will ensure that it will never be deallocated */
    1   /* is clean */
};

/*
 * Allocate an array of size 'n'.
 * Observe that the overhead is not guarateed.
 */
struct vector *allocate_array_with_overhead(int n,int overhead)
{
  int i;
  struct vector *p;

  if (!batch_mode && (n < 0 || n > MAX_ARRAY_SIZE))
    error("Illegal array size.\n",0,0,0,0,0,0,0,0,0);

  if (n == 0)
  {
    p = &null_vector;
    p->ref++;
  }else{
    num_arrays++;
    total_array_size += VECTOR_SIZE(n);

    /* overhead stuff */
    i=n+overhead;
    if(i>MAX_ARRAY_SIZE && !batch_mode) i=MAX_ARRAY_SIZE;
    p = ALLOC_VECTOR(i);
    p->malloced_size=i;
    p->ref = 1;
    p->size = n;
    p->flags=0;
    p->types=-1;
    if(current_object && current_object->user)
      p->user = current_object->user;
    else
      p->user=0;

    if (p->user)
      p->user->size_array += n;
    for (i=0; i<n; i++)
      p->item[i] = const0;
  }
  return p;
}

struct vector *allocate_array(int n)
{
  return allocate_array_with_overhead(n,0);
}

void free_vector(struct vector *p)
{
    int i;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    p->ref--;
#ifdef DEBUG
    if(p->size>p->malloced_size)
    {
      fatal("Impossible array.",0,0,0,0,0,0,0,0,0);
    }
    if(p->ref<0) fatal("Freeing array too many times.\n",0,0,0,0,0,0,0,0,0);
#endif
    if (p->ref > 0)
	return;
#ifdef DEBUG
    if (p == &null_vector)
	fatal("Tried to free the zero-size shared vector.\n",0,0,0,0,0,0,0,0,0);
#endif
    remove_array_from_stack(p);
    for (i=0; i<p->size; i++)
	free_svalue(&p->item[i]);
    if (p->user)
	p->user->size_array -= p->size;
    num_arrays--;
    total_array_size -= sizeof (struct vector) + sizeof (struct svalue) *
	(p->size-1);
    free((char *)p);
}

/* OBS, destructive */
struct vector *resize_array(struct vector *p,int n)
{
  struct vector *res;
  int e;
  if(p->size==n) return p;
  if(n>p->size)
  {
    /* we should grow the array */
    if(n<=p->malloced_size)
    {
      /* it fits */
      p->types=-1;
      for(e=p->size+1;e<n;e++) p->item[e]=const0;
      e=n-p->size;
      total_array_size +=sizeof(struct svalue) *e ;
      if(p->user) p->user->size_array+=e;
      p->size=n;
      return p;
    }
    res=allocate_array_with_overhead(n,n>>4+1);
    for(e=0;e<p->size;e++) assign_svalue_no_free(res->item+e,p->item+e);
    for(;e<n;e++) res->item[e]=const0;
    free_vector(p);
    return res;
  }else{
    /* we should shrink the array */
    if(n<=p->malloced_size>>11)
    {
      /* overhead to big, reallocate the array  */
      res = slice_array(p, 0, n-1);
      free_vector(p);
      return res;
    }
    e=n-p->size;
    total_array_size += sizeof (struct svalue) * e;
    if (p->user) p->user->size_array += e;
    p->size = n;
    return p;
  }
}

struct vector *shrink_array(struct vector *p,int n)
{
  return resize_array(p,n);
}

/* Rewritten by Profezzorn */

/* a binary string capable explode */

struct vector *explode(struct svalue *str,struct svalue *del)
{
  int str_len,del_len,e,tmp,begin;
  struct vector *v;
  char *s,*d;

  s=str->u.string;
  d=del->u.string;

  str_len=my_strlen(str);
  del_len=my_strlen(del);

  if(!del_len)
  {
    v=allocate_array(str_len);
    for(e=0;e<str_len;e++)
    {
      v->item[e].type=T_STRING;
      v->item[e].subtype=STRING_SHARED;
      v->item[e].u.string=make_shared_binary_string(s+e,1);
    }
  }else{
    for(tmp=e=0;e<=str_len-del_len;)
    {
      if(!memcmp(s+e,d,del_len))
      {
        tmp++;
        e+=del_len;
      }else{
        e++;
      }
    }

    v=allocate_array(tmp+1);

    for(begin=tmp=e=0;e<=str_len-del_len;)
    {
      if(!memcmp(s+e,d,del_len))
      {
        v->item[tmp].type=T_STRING;
        v->item[tmp].subtype=STRING_SHARED;
        v->item[tmp].u.string=make_shared_binary_string(s+begin,e-begin);
        tmp++;
        e+=del_len;
        begin=e;
      }else{
        e++;
      }
    }
    v->item[tmp].type=T_STRING;
    v->item[tmp].subtype=STRING_SHARED;
    v->item[tmp].u.string=make_shared_binary_string(s+begin,str_len-begin);
  }
  v->types=T_STRING; /* nothing but us strings here */
  return v;
}

/* a binary implode */

void implode(struct svalue *to,struct vector *v,struct svalue *del)
{
  int del_len,len,e;
  char *s,*s2,*del_str;
  
  del_len=my_strlen(del);
  del_str=del->u.string;

  len=del_len*(v->size-1);
  for(e=0;e<v->size;e++)
  {
    if(v->item[e].type!=T_STRING) continue;
    len+=my_strlen(v->item+e);
  }

  if(len<=0)
  {
    to->type=T_STRING;
    to->subtype=STRING_CONSTANT;
    to->u.string="";
  }else{
    s2=s=begin_shared_string(len);
    
    for(e=0;e<v->size;e++)
    {
      if(e)
      {
        memcpy(s2,del_str,del_len);
        s2+=del_len;
      }
      if(v->item[e].type!=T_STRING) continue;

      len=my_strlen(v->item+e);
      memcpy(s2,v->item[e].u.string,len);
      s2+=len;
    }
    s=end_shared_string(s); /* may or may not use same area */
    to->type=T_STRING;
    to->subtype=STRING_SHARED;
    to->u.string=s;
  }
}


/* The following two functions are for the replace() efun....
 *   Profezzorn
 */

struct vector *array_replace(struct vector *from,struct svalue *what,struct svalue *to)
{
  int e;
  struct vector *ret;
  ret=allocate_array(from->size);
  for(e=0;e<from->size;e++)
  {
    if(is_eq(what,&(from->item[e])))
      assign_svalue_no_free(&(ret->item[e]),to);
    else
      assign_svalue_no_free(&(ret->item[e]),&(from->item[e]));
  }
  return ret;
}

/*
 * Slice of an array.
 */
struct vector *slice_array(struct vector *p,int from,int to)
{
    struct vector *d;
    int cnt;
    
    if (from < 0)
	from = 0;
    if (from >= p->size)
	return allocate_array(0); /* Slice starts above array */
    if (to >= p->size)
	to = p->size-1;
    if (to < from)
	return allocate_array(0); 
    
    d = allocate_array(to-from+1);
    for (cnt=from;cnt<=to;cnt++) 
	assign_svalue_no_free (&d->item[cnt-from], &p->item[cnt]);
    
    return d;
}

/* EFUN: filter_array
   
   Runs all elements of an array through ob->func()
   and returns an array holding those elements that ob->func
   returned 1 for.
   Hacking a little: if ob==0 then we call in the object in the array
                     instead.......
   */
struct vector *filter(p, lambda, extra, num_extra_arg)
  struct vector *p;
  struct svalue *lambda;
  struct svalue *extra;
  int num_extra_arg;
{
  struct vector *v,*r;
  int e,num;

  if (p->size<1) return allocate_array(0);
  v=map_array(p,lambda,extra,num_extra_arg);
  for(num=e=0;e<v->size;e++)
    if(v->item[e].type!=T_NUMBER || (v->item[e].u.number!=0)) num++;

  r=allocate_array(num);
  num=0;
  for(e=0;e<v->size;e++)
    if(v->item[e].type!=T_NUMBER || (v->item[e].u.number!=0))
      assign_svalue_no_free(&r->item[num++], &p->item[e]);
  free_vector(v);
  return r;
}

/* Unique maker
   
   These routines takes an array of objects and calls the function 'func'
   in them. The return values are used to decide which of the objects are
   unique. Then an array on the below form are returned:
   
   ({
   ({Same1:1, Same1:2, Same1:3, .... Same1:N }),
   ({Same2:1, Same2:2, Same2:3, .... Same2:N }),
   ({Same3:1, Same3:2, Same3:3, .... Same3:N }),
   ....
   ....
   ({SameM:1, SameM:2, SameM:3, .... SameM:N }),
   })
   i.e an array of arrays consisting of lists of objectpointers
   to all the nonunique objects for each unique set of objects.
   
   The basic purpose of this routine is to speed up the preparing of the
   array used for describing.
   
   */

struct unique
{
    int count;
    struct svalue *val;
    struct svalue mark;
    struct unique *same;
    struct unique *next;
};


INLINE int sameval(struct svalue *arg1,struct svalue *arg2)
{
  if (!arg1 || !arg2) return 0;
  return is_eq(arg1,arg2);
}


static int put_in(ulist,marker,elem)
    struct unique **ulist;
    struct svalue *marker;
    struct svalue *elem;
{
    struct unique *llink,*slink,*tlink;
    int cnt,fixed;
    
    llink= *ulist;
    cnt=0; fixed=0;
    while (llink) {
	if ((!fixed) && (sameval(marker,&(llink->mark)))) {
	    for (tlink=llink;tlink->same;tlink=tlink->same) (tlink->count)++;
	    (tlink->count)++;
	    slink = (struct unique *) xalloc(sizeof(struct unique));
	    slink->count = 1;
	    assign_svalue_no_free(&slink->mark,marker);
	    slink->val = elem;
	    slink->same = 0;
	    slink->next = 0;
	    tlink->same = slink;
	    fixed=1; /* We want the size of the list so do not break here */
	}
	llink=llink->next; cnt++;
    }
    if (fixed) return cnt;
    llink = (struct unique *) xalloc(sizeof(struct unique));
    llink->count = 1;
    assign_svalue_no_free(&llink->mark,marker);
    llink->val = elem;
    llink->same = 0;
    llink->next = *ulist;
    *ulist = llink;
    return cnt+1;
}


struct vector *
make_unique(arr,func,skipnum)
    struct vector *arr;
    char *func;
    struct svalue *skipnum;
{
    struct svalue *v;
    struct vector *res,*ret;
    struct unique *head,*nxt,*nxt2;
    
    int cnt,ant,cnt2;
    
    if (arr->size < 1)
	return allocate_array(0);

    head = 0; ant=0; arr->ref++;
    for(cnt=0;cnt<arr->size;cnt++) if (arr->item[cnt].type == T_OBJECT) {
	v = apply(func,arr->item[cnt].u.ob,0);
	if ((!v) || (v->type!=skipnum->type) || !sameval(v,skipnum)) 
	    if (v) ant=put_in(&head,v,&(arr->item[cnt]));
    }
    arr->ref--;
    ret = allocate_array(ant);
    
    for (cnt=ant-1;cnt>=0;cnt--) { /* Reverse to compensate put_in */
	ret->item[cnt].type = T_POINTER;
	ret->item[cnt].u.vec = res = allocate_array(head->count);
	nxt2 = head;
	head = head->next;
	cnt2 = 0;
	while (nxt2) {
	    assign_svalue_no_free (&res->item[cnt2++], nxt2->val);
	    free_svalue(&nxt2->mark);
	    nxt = nxt2->same;
	    free((char *) nxt2);
	    nxt2 = nxt;
	}
	if (!head) 
	    break; /* It shouldn't but, to avoid skydive just in case */
    }
    return ret;
}

/*
 * End of Unique maker
 *************************
 */

/* Concatenation of two arrays into one
 */
struct vector *add_array(p,r)
    struct vector *p, *r;
{
    int cnt,res;
    struct vector *d;
    
    d = allocate_array(p->size+r->size);
    res=0;
    for (cnt=0;cnt<p->size;cnt++) {
	assign_svalue_no_free (&d->item[res],&p->item[cnt]);
	res++; 
    }
    for (cnt=0;cnt<r->size;cnt++) {
	assign_svalue_no_free (&d->item[res],&r->item[cnt]);
	res++; 
    }
    return d;
}

struct vector *subtract_array(struct vector *minuend,
			      struct vector * subtrahend)
{
  struct vector *vtmp;
  struct vector *difference;
  struct svalue *source,*dest;
  int i;

  vtmp=allocate_array(1);
  vtmp->item[0].type=T_ALIST_PART;
  if(subtrahend->ref==1)
  {
    vtmp->item[0].u.vec = subtrahend;
    subtrahend->ref++;
  }else{
    vtmp->item[0].u.vec = slice_array(subtrahend,0,subtrahend->size);
  }
  order_alist(vtmp);
  subtrahend = vtmp->item[0].u.vec;
  difference = allocate_array(minuend->size);
  for (source = minuend->item, dest = difference->item, i = minuend->size;
       i--; source++)
  {

    int assoc PROT((struct svalue *key, struct vector *));
/*
     if ((source->type & (T_OBJECT | T_FUNCTION)) &&
	source->u.ob->flags & O_DESTRUCTED)

      assign_svalue(source, &const0);
*/
    if ( assoc(source, subtrahend) < 0 )
      assign_svalue_no_free(dest++, source);
  }
  free_vector(vtmp);
  return shrink_array(difference, dest-difference->item);
}


/* Returns an array of all objects contained in 'ob'
 */
struct vector *all_inventory(struct object *ob)
{
  struct vector *d;
  struct object *cur;
  int cnt,res;
    
  cnt=0;
  for (cur=ob->contains; cur; cur = cur->next_inv)
    cnt++;
    
  if (!cnt)
    return allocate_array(0);

  d = allocate_array(cnt);
  cur=ob->contains;
    
  for (res=0;res<cnt;res++) {
    d->item[res].type=T_OBJECT;
    d->item[res].u.ob = cur;
    add_ref(cur,"all_inventory");
    cur=cur->next_inv;
  }
  return d;
}


/* Runs all elements of an array through ob::func
   and replaces each value in arr by the value returned by ob::func
   Hacking a little: if ob==0 then we call in the object in the array
                     instead.......
   */
struct vector *map_array(
	struct vector *p,
	struct svalue *lambda,
	struct svalue *extra,
	int num_extra_arg)
{
  struct vector *r;
  struct svalue *v;
  char *func=0; /* make gcc happy */
  int cnt,no;
    
  r=0;
  if(lambda->type==T_STRING)
  {
    func=lambda->u.string;
  }else if(lambda->type==T_FUNCTION){
    if(lambda->u.ob->flags & O_DESTRUCTED)
    {
      error("Invalid lambda to map_array.\n",0,0,0,0,0,0,0,0,0);
    }
  }else if(lambda->type==T_NUMBER){
    /* no action here */
  }else{
    error("Bad arg to map_array.\n",0,0,0,0,0,0,0,0,0);
  }

  if (p->size<1) return allocate_array(0);
 
  r = allocate_array(p->size);
 
  for (cnt=0;cnt<p->size;cnt++)
  {
    check_eval_cost();
    switch(lambda->type)
    {
      case T_NUMBER:
        if(p->item[cnt].type!=T_FUNCTION ||
          (p->item[cnt].u.ob->flags & O_DESTRUCTED)) continue;
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply_lambda(p->item+cnt,no);
        break;

      case T_STRING:
        if(p->item[cnt].type!=T_OBJECT ||
          (p->item[cnt].u.ob->flags & O_DESTRUCTED)) continue;
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply(func,p->item[cnt].u.ob,no);
        break;

      case T_FUNCTION:
        if(lambda->u.ob->flags & O_DESTRUCTED)
          error("object used by map_array destructed\n",0,0,0,0,0,0,0,0,0);
        push_svalue(&p->item[cnt]);
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply_lambda(lambda,no+1);
        break;

      default:
	fatal("Fatal error in map_array.\n",0,0,0,0,0,0,0,0,0);
	return 0;
    }
    if(v) assign_svalue_no_free(&r->item[cnt], v);
  }
  return r;
}

int search_array(
	struct vector *p,
	struct svalue *lambda,
	struct svalue *extra,
	int num_extra_arg)
{
  struct svalue *v;
  char *func=0;  /* make gcc happy */
  int cnt,no;
    
  if(lambda->type==T_STRING)
  {
    func=lambda->u.string;
  }else if(lambda->type==T_FUNCTION){
    if(lambda->u.ob->flags & O_DESTRUCTED)
    {
      error("Invalid lambda to map_array.\n",0,0,0,0,0,0,0,0,0);
    }
  }else if(lambda->type==T_NUMBER){
    /* no action here */
  }else{
    error("Bad arg to map_array.\n",0,0,0,0,0,0,0,0,0);
  }

  if (p->size<1) return -1;
 
  for (cnt=0;cnt<p->size;cnt++)
  {
    check_eval_cost();
    switch(lambda->type)
    {
      case T_NUMBER:
        if(p->item[cnt].type!=T_FUNCTION ||
          (p->item[cnt].u.ob->flags & O_DESTRUCTED)) continue;
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply_lambda(p->item+cnt,no);
        break;

      case T_STRING:
        if(p->item[cnt].type!=T_OBJECT ||
          (p->item[cnt].u.ob->flags & O_DESTRUCTED)) continue;
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply(func,p->item[cnt].u.ob,no);
        break;

      case T_FUNCTION:
        if(lambda->u.ob->flags & O_DESTRUCTED)
          error("object used by map_array destructed\n",0,0,0,0,0,0,0,0,0);
        push_svalue(&p->item[cnt]);
        for(no=0;no<num_extra_arg;no++) push_svalue(extra+no);
        v=apply_lambda(lambda,no+1);
        break;

      default:
	fatal("Fatal error in search_array.\n",0,0,0,0,0,0,0,0,0);
	return 0;
    }
    if(v && (v->type!=T_NUMBER || v->u.number==0)) return cnt;
  }
  return -1;
}

static struct svalue *func;
static struct svalue *extra_arg;
static int num_extra_arg;

static int sort_array_cmp(struct svalue *p1, struct svalue *p2)
{
  struct svalue *d;
  int e;
  if(!func || func->type!=T_FUNCTION)
  {
    if(p1->type>p2->type) return 1;
    return is_gt(p1,p2);
  }

  if (func->u.ob->flags & O_DESTRUCTED)
    error("object used by sort_array destructed",0,0,0,0,0,0,0,0,0);

  push_svalue(p1);
  push_svalue(p2);
  for(e=0;e<num_extra_arg;e++)
  {
    push_svalue(extra_arg+e);
  }
  d = apply_lambda(func,num_extra_arg+2);
  if (!d) return 0;
  if(d->type!=T_NUMBER) return 1;
  return d->u.number;
}

struct vector *sort_array(
	struct vector *v,
	struct svalue *fun,
	struct svalue *extra,
	int num)
{
    struct vector *outlist;
    struct svalue *root,*out;
    int n, e;

    func=fun;
    extra_arg=extra;
    num_extra_arg=num;

    n=v->size;
    outlist = allocate_array(n);
    root = v->item;
    out = outlist->item;
    for(e=0;e<n;e++,out++,root++)
    {
      if((root->type & (T_OBJECT|T_FUNCTION)) && root->u.ob->flags & O_DESTRUCTED)
	continue;
      assign_svalue_no_free(out,root);
    }
    msort(outlist->item,n,sizeof(struct svalue),sort_array_cmp);
    return outlist;
}

/* Turns a structured array of elements into a flat array of elements.
   */
static int num_elems(struct vector *arr)
{
  int cnt,il;

  cnt = arr->size;

  for (il=0;il<arr->size;il++) 
    if (arr->item[il].type == T_POINTER) 
      cnt += num_elems(arr->item[il].u.vec) - 1;
  return cnt;
}

struct svalue *splatt(struct vector *from,struct svalue *to)
{
  struct svalue *src;
  int no;

  src=from->item;
  no=from->size;

  for(;no<=0;src++,no--)
  {
    if(src->type!=T_POINTER)
      assign_svalue(to++,src);
    else
      to=splatt(src->u.vec,to);
  }
  return to;
}

struct vector *flatten_array(struct vector *arr)
{
  int max;
  struct vector *res;
    
  if (arr->size < 1) 
    return allocate_array(0);

  max = num_elems(arr);

  if (arr->size == max)
    return arr;

  if (max == 0)			/* In the case arr is an array of empty arrays */
    return allocate_array(0);

  res = allocate_array(max); 
  (void)splatt(arr,res->item);
  return res;
}
/*
 * deep_inventory()
 *
 * This function returns the recursive inventory of an object. The returned 
 * array of objects is flat, ie there is no structure reflecting the 
 * internal containment relations.
 *
 */
int number_deep_inv(struct object *ob)
{
  int num;
  num=1;
  for(ob=ob->contains;ob;ob=ob->next_inv)
    num+=number_deep_inv(ob);
  return num;
}

/* This one is faster, but the array is even more scrambled compared to
 * the normal deep_inventory() /Profezzorn
 */
struct vector *deep_inventory(struct object *ob,int take_top)
{
  int num;
  struct vector *res;
  struct svalue *curr,*next;
  num=number_deep_inv(ob);
  if(take_top)
  {
    res=allocate_array(num);
    add_ref(ob,"deep_inventory");
    res->item[0].type=T_OBJECT;
    res->item[0].u.ob=ob;
    next=curr=res->item+1;
  }else{
    res=allocate_array(num-1);
    curr=next=res->item;
  }
  for(;curr<=next;curr++)
  {
    for(ob=ob->contains;ob;ob=ob->next_inv)
    {
      add_ref(ob,"deep_inventory");
      next->type=T_OBJECT;
      next->u.ob=ob;
      next++;
    }
    ob=curr->u.ob;
  }
  return res;
}

INLINE int alist_cmp(struct svalue *p1,struct svalue *p2)
{
    register int d;

    if (d = p1->type - p2->type) return d;
    switch(p1->type)
    {
      case T_FLOAT:
        if(p1->u.fnum>p2->u.fnum) return 1;
        if(p1->u.fnum<p2->u.fnum) return -1;
        return 0;

      case T_FUNCTION:
        if (d = p1->u.number - p2->u.number) return d;
        if (d = p1->subtype - p2->subtype) return d;
        return 0;

      default:
        if(p1->u.number>p2->u.number) return 1;
        if(p1->u.number<p2->u.number) return -1;
        return 0;

    }
}

#if 1
struct vector *globber;

int new_alist_cmp(int *a,int *b)
{
  return alist_cmp(globber->item+*a,globber->item+*b);
}

void order_alist(struct vector *inlist)
{
  int *tmp;
  struct vector *tmplist,*tmptmplist;
  int e,d;
  globber=inlist->item[0].u.vec;
  if(!globber->size) return;
  tmp=(int *)malloc(globber->size*sizeof(int));

  for(e=0;e<globber->size;e++)
  {
    if(globber->item[e].type==T_STRING &&
	 globber->item[e].subtype!=STRING_SHARED)
    {
      char *str = make_shared_string(globber->item[e].u.string);
      free_svalue(globber->item+e);
      globber->item[e].type = T_STRING;
      globber->item[e].subtype = STRING_SHARED;
      globber->item[e].u.string = str;
    }
    tmp[e]=e;
  }
  fsort(tmp,globber->size,sizeof(int),new_alist_cmp);

  tmplist=allocate_array(globber->size);
  for(e=0;e<inlist->size;e++)
  {
    if(inlist->item[e].type==T_ALIST_PART)
    {
      tmptmplist=inlist->item[e].u.vec;
      for(d=0;d<tmplist->size;d++) tmplist->item[d]=tmptmplist->item[tmp[d]];
      inlist->item[e].u.vec=tmplist;
      tmplist=tmptmplist;
    }
  }
  /* for(e=0;e<tmplist->size;e++) tmplist->item[e].type=T_INVALID; */
  tmplist->size=0; /* instead of setting all the items in it to const0 */
  free_vector(tmplist);
  free((char *)tmp);
}

#else

#define MIN(X,Y) ((X)<(Y)?(X):(Y))
#define MAX(X,Y) ((X)>(Y)?(X):(Y))
#define START_ARRAY_SIZE 16
#define VECTORS 16

typedef int *intp;
static struct vector *index_vector;
static intp vectors[VECTORS];
static int sizes[VECTORS];

typedef struct vector * vectorp;

static void low_radix(int *v,int len,int shift)
{
  int indexes[VECTORS];
  int e,d,tmp;

  if(!len) return;

  for(e=0;e<VECTORS;e++) indexes[e]=0;

  for(e=0;e<len;e++)
  {
    tmp=index_vector->item[v[e]].u.number ^ 0x80000000;
    tmp=( tmp >> shift ) & (VECTORS-1);

    if(indexes[tmp]==sizes[tmp])
    {
      sizes[tmp]*=2;
      vectors[tmp]=(intp)realloc((char *)vectors[tmp],sizeof(int)*sizes[tmp]);
    }
    
    vectors[tmp][indexes[tmp]++]=v[e];
  }

  tmp=0;
  for(e=0;e<VECTORS;e++)
    for(d=0;d<indexes[e];d++)
      v[tmp++]=vectors[e][d];

  if(shift)
  {
    shift-=4;
    tmp=0;
    for(e=0;e<VECTORS;e++)
    {
      low_radix(v+tmp,indexes[e],shift);
      tmp+=indexes[e];
    }
  }
}

int float_cmp(int *a,int *b)
{
  if(index_vector->item[*a].u.fnum<index_vector->item[*b].u.fnum)
    return -1;
  return 0;
}

void radix(int *v,int len)
{
  int indexes[VECTORS];
  int e,d,tmp;

  if(!len) return;

  tmp=MIN(len,START_ARRAY_SIZE);
  for(e=0;e<VECTORS;e++)
  {
    vectors[e]=(intp)malloc(sizeof(int)*tmp);
    sizes[e]=tmp;
    indexes[e]=0;
  }

  for(e=0;e<len;e++)
  {
    d=index_vector->item[v[e]].type;
    tmp=0;
    if(d & 0xff00) tmp+=8;
    if(d & 0xf0f0) tmp+=4;
    if(d & 0xcccc) tmp+=2;
    if(d & 0xaaaa) tmp+=1;

    if(indexes[tmp]==sizes[tmp])
    {
      sizes[tmp]*=2;
      vectors[tmp]=(intp)realloc((char *)vectors[tmp],sizeof(int)*sizes[tmp]);
    }
    
    vectors[tmp][indexes[tmp]++]=v[e];
  }

  tmp=0;
  for(e=0;e<VECTORS;e++)
    for(d=0;d<indexes[e];d++)
      v[tmp++]=vectors[e][d];

  tmp=0;
  for(e=0;e<VECTORS;e++)
  {
    if((1<<e)==T_FLOAT)
    {
      if(indexes[e])
        fsort(v+tmp,indexes[e],sizeof(int),float_cmp);
    }else{
      low_radix(v+tmp,indexes[e],28);
    }
    tmp+=indexes[e];
  }

  for(e=0;e<VECTORS;e++)
    free((char *)vectors[e]);
}

void order_alist(struct vector *v)
{
  int *a;
  int size,e,d;
  struct vector *tmp,*tmp2;

  index_vector=v->item[0].u.vec;
  size=index_vector->size;
  a=(int *)malloc(sizeof(int)*size);

  for(e=0;e<size;e++) a[e]=e;

  radix(a,size);

  tmp=allocate_array(size);
  for(e=0;e<v->size;e++)
  {
    if(v->item[e].type!=T_ALIST_PART) continue;
    tmp2=v->item[e].u.vec;
    for(d=0;d<size;d++)
      tmp->item[d]=tmp2->item[a[d]];
    v->item[e].u.vec=tmp;
    tmp=tmp2;
  }
  tmp->size=0; /* instead of setting all the items in it to const0 */
  free_vector(tmp);

}

#endif
int search_alist(struct svalue *key,struct vector * keylist)
{
    int i, o, d;

    if (!keylist->size) return 0;
    i = (keylist->size) >> 1;
    o = (i+2) >> 1;
    for (;;) {
	d = alist_cmp(key, &keylist->item[i]);
	if (d<0) {
	    i -= o;
	    if (i<0) {
		i = 0;
	    }
	} else if (d>0) {
	    i += o;
	    if (i >= keylist->size) {
	        i = keylist->size-1;
	    }
	} else {
	    return i;
	}
	if (o<=1) {
	    if (alist_cmp(key, &keylist->item[i]) > 0) return i+1;
	    return i;
	}
	o = (o+1) >> 1;
    }
}

/* obs, destructive /Profezzorn */

struct svalue stmp = { T_NUMBER } ;
struct vector *insert_alist(
	struct svalue *key,
	struct svalue *key_data,
	struct vector *list)
{
  int i,j,ix;
  int keynum;
  int insert;

  if(stmp.type!=T_NUMBER) free_svalue(&stmp);

  if (key->type == T_STRING && key->subtype != STRING_SHARED)
  {
    stmp.u.string = make_shared_string(key->u.string);
    stmp.type=T_STRING;
    stmp.subtype=STRING_SHARED;
    key = &stmp;
  }
  keynum = list->item[0].u.vec->size;
  ix = search_alist(key,list->item[0].u.vec);
  insert=ix==keynum || alist_cmp(key, &list->item[0].u.vec->item[ix]);
  for (i=0; i < list->size; i++)
  {
    if(list->item[i].type==T_ALIST_PART)
    {
      struct vector *tmp = list->item[i].u.vec;
      struct svalue *sv;
#ifdef DEBUG
      if(tmp->size!=keynum)
	fatal("If this is an alist, then I'll be buggered.\n",0,0,0,0,0,0,0,0,0);
#endif
      if(insert)
      {
    
	tmp = resize_array(tmp,keynum+1);
        for (j=keynum; j > ix; j--)
          tmp->item[j]=tmp->item[j-1];

	tmp->item[ix]=const0;
	sv= i ? key_data+i-1 : key;
	if(sv->type | T_VECTOR)
	  if(check_for_circularity(tmp,sv->u.vec))
	    error("Trying to make circular alist.\n",0,0,0,0,0,0,0,0,0);

        assign_svalue_no_free(tmp->item+ix,sv);
	tmp->types|=sv-> type;;
        list->item[i].u.vec=tmp;
      }else{
        if(i)
	{
	  sv= key_data+i-1;
	  if(sv->type | T_VECTOR)
	    if(check_for_circularity(tmp,sv->u.vec))
	      error("Trying to make circular alist.\n",0,0,0,0,0,0,0,0,0);

          assign_svalue(tmp->item+ix, sv);
	  tmp->types|=sv->type;
	}
      }
    }
  }
#ifdef DEBUG
  for(i=0;i<list->size;i++)
  {
    if(list->item[i].type==T_ALIST_PART && list->item[0].u.vec->size!=
       list->item[i].u.vec->size)
      fatal("Hmm, this is _not_ what I had in mind....\n",0,0,0,0,0,0,0,0,0);
  }
#endif
  if(stmp.type!=T_NUMBER) free_svalue(&stmp);
  return list;
}

int assoc(struct svalue *key,struct vector * list)
{
  int i;
  extern char* findstring PROT((char*));

  if (key->type == T_STRING && key->subtype != STRING_SHARED) {
    static struct svalue stmp = { T_STRING, STRING_SHARED };

    stmp.u.string = findstring(key->u.string);
    if (!stmp.u.string) return -1;
    key = &stmp;
  }
  i = search_alist(key, list);
  if (i == list->size || alist_cmp(key, &list->item[i]))
    i = -1;
  return i;
}

struct vector *intersect_alist(struct vector *a1,struct vector * a2)
{
  struct vector *a3;
  int d, l, i1, i2, a1s, a2s;

  a1s = a1->size;
  a2s = a2->size;
  a3 = allocate_array( a1s < a2s ? a1s : a2s);
  for (i1=i2=l=0; i1 < a1s && i2 < a2s; ) {
    d = alist_cmp(&a1->item[i1], &a2->item[i2]);
    if (d<0)
    {
      i1++;
    }else if (d>0){
      i2++;
    }else {
      assign_svalue_no_free(&a3->item[l++], &a2->item[i1++,i2++]);
    }
  }
  return shrink_array(a3, l);
}

struct vector *symmetric_difference_alist(struct vector *a1,struct vector *a2)
{
  struct vector *a3;
  int d, l, i1, i2, a1s, a2s;

  a1s = a1->size;
  a2s = a2->size;
  a3 = allocate_array( a1s + a2s );
  for (i1=i2=l=0; i1 < a1s && i2 < a2s; ) {
    d = alist_cmp(&a1->item[i1], &a2->item[i2]);
    if (d<0)
      assign_svalue_no_free(&a3->item[l++], &a1->item[i1++]);
    else if (d>0)
      assign_svalue_no_free(&a3->item[l++], &a2->item[i2++]);
    else {
      i1++;
      i2++;
    }
  }
  while (i1 < a1s )
    assign_svalue_no_free(&a3->item[l++], &a1->item[i1++]);
  while (i2 < a2s )
    assign_svalue_no_free(&a3->item[l++], &a2->item[i2++]);
  return shrink_array(a3, l);
}

struct vector *match_regexp(struct vector *v,char * pattern)
{
  struct regexp *reg;
  char *res;
  int i, num_match;
  struct vector *ret;
  extern int eval_cost;

  if (v->size == 0)
    return allocate_array(0);
  reg = regcomp(pattern, 0);
  if (reg == 0)
    return 0;
  res = (char *)allocaX(v->size);
  for (num_match=i=0; i < v->size; i++) {
    res[i] = 0;
    if (v->item[i].type != T_STRING)
      continue;
    eval_cost++;
    if (regexec(reg, v->item[i].u.string) == 0)
      continue;
    res[i] = 1;
    num_match++;
  }
  ret = allocate_array(num_match);
  for (num_match=i=0; i < v->size; i++) {
    if (res[i] == 0)
      continue;
    assign_svalue_no_free(&ret->item[num_match], &v->item[i]);
    num_match++;
  }
  free((char *)reg);
  return ret;
}

/*
 * Returns a list of all inherited files.
 *
 * Must be fixed so that any number of files can be returned, now max 256
 * (Sounds like a contradiction to me /Lars).
 * Now it can return any number of files. /Profezzorn
 */
struct vector *inherit_list(struct object *ob)
{
  struct vector *ret;
  struct program *pr;
  int e;

  pr=ob->prog;
  ret=allocate_array(pr->num_inherited);
  for(e=0;e<pr->num_inherited;e++)
  {
    ret->item[e].type=T_STRING;
    ret->item[e].subtype=STRING_SHARED;
    ret->item[e].u.string=copy_shared_string(pr->inherit[e].prog->name);
  }
  return ret;
}

/*
 * When a vector is given as argument to an efun, all items has to be
 * checked if there would be an destructed object.
 * A bad problem currently is that a vector can contain another vector, so this
 * should be tested too. But, there is currently no prevention against
 * recursive vectors, which means that this can not be tested. Thus, the game
 * may crash if a vector contains a vector that contains a destructed object
 * and this top-most vector is used as an argument to an efun.
 */
/* The game won't crash when doing simple operations like assign_svalue
 * on a destructed object. You have to watch out, of course, that you don't
 * apply a function to it.
 * to save space it is preferable that destructed objects are freed soon.
 *   amylaar
 */

void donk_alist_item(struct vector *m,int i)
{
  int e,d,s;
  for(s=e=0;e<m->size;e++)
  {
    if(m->item[e].type!=T_ALIST_PART) continue;
    free_svalue(m->item[e].u.vec->item+i);
    for(d=i+1;d<m->item[e].u.vec->size;d++)
      m->item[e].u.vec->item[d-1]=m->item[e].u.vec->item[d];
    m->item[e].u.vec->size--;
    s++;
  }
  total_array_size -= s*sizeof (struct svalue);
  if (m->user) m->user->size_array-=s;
}

/* I'm a big beleiver in lazy evaluation algorithms */

void check_vector_for_destruct(struct vector *v)
{
  int types,e;
  /* if there are no objects or function, why check? */
  if(!(v->types & (T_OBJECT | T_FUNCTION)))
    return;

  types=0;
  for(e=v->size-1;e>=0;e--)
  {
    if(v->item[e].type & (T_OBJECT | T_FUNCTION))
    {
      if(v->item[e].u.ob->flags & O_DESTRUCTED)
      {
	short tmp;
	tmp=v->item[e].type=T_OBJECT;
        free_svalue(v->item+e);
	v->item[e].subtype=tmp?NUMBER_DESTRUCTED_OBJECT:NUMBER_DESTRUCTED_FUNCTION;
      }
    }

    types|=v->item[e].type;
  }
  v->types=types;
}

void check_alist_for_destruct(struct vector *v)
{
  int e,types;
  struct vector *w;
  check_vector_for_destruct(v);
  w=v->item[0].u.vec;

  if(w->types & (T_OBJECT | T_FUNCTION))
  {
    types=0;
    for(e=0;e<w->size;e++)
    {
      if((w->item[e].type & (T_OBJECT | T_FUNCTION)) &&
        (w->item[e].u.ob->flags & O_DESTRUCTED))
      {
        donk_alist_item(v,e);
        e--;
      }else{
        types|=w->item[e].type;
      }
    }
    w->types=types;
  }
  for(e=1;e<v->size;e++)
    if(v->item[e].type & T_ALIST_PART)
      check_vector_for_destruct(v->item[e].u.vec);
}

/* check if v2 is in v */
int check_for_circularity(struct vector *v,struct vector *v2)
{
  int types,e;
  /* if there are no vectors, why search? */
  if(!(v->types & T_VECTOR))
    return 0;


  if(v==v2) return 1;

  types=0;
  for(e=v->size-1;e>=0;e--)
  {
    if(v->item[e].type & T_VECTOR)
      if(check_for_circularity(v->item[e].u.vec,v2))
        return 1;

    types|=v->item[e].type;
  }

  /* if we reach this, types is updated.
   * as recursive arrays gives errors, we should come here
   * almost every time.
   */
  v->types=types;
  return 0;
}

struct vector *slice_alist(struct vector *m,int aa,int bb)
{
  int e;
  struct vector *v;
  v=allocate_array(m->size);
  for(e=0;e<m->size;e++)
  {
    if(m->item[e].type==T_ALIST_PART)
    {
      v->item[e].type=T_ALIST_PART;
      v->item[e].u.vec=slice_array(m->item[e].u.vec,aa,bb);
    }else{
      assign_svalue(v->item+e,m->item+e);
    }
  }
  return v;
}

struct vector *alist_range(struct vector *m,struct svalue *a,struct svalue *b)
{
  int aa,bb;
  int e;
  struct vector *v;
  v=m->item[0].u.vec;
  aa=search_alist(a, v);
  bb=search_alist(b, v);
  if(aa>bb) { e=aa; aa=bb; bb=e; }

  while(aa>0 && 
     v->item[aa].u.number==v->item[aa-1].u.number &&
     v->item[aa].type==v->item[aa-1].type)
	aa--;

  while(bb<v->size-1 && 
     v->item[bb].u.number==v->item[bb+1].u.number &&
     v->item[bb].type==v->item[bb+1].type)
	bb++;

  return slice_alist(m,aa,bb);
}

struct vector *allocate_n_array(struct svalue *sp,int t)
{
  struct vector *v;
  int e;
  extern int eval_cost;
  if(sp->type!=T_NUMBER)
    error("Non-numeric array dimension.\n",0,0,0,0,0,0,0,0,0);
  v=allocate_array(sp->u.number);
  eval_cost+=sp->u.number*AVERAGE_COST;
  t--;
  v->types=T_NUMBER;
  if(!t) return v;
  sp++;
  v->types=T_POINTER;
  for(e=0;e<v->size && eval_cost<MAX_COST;e++)
  {
    v->item[e].type=T_POINTER;
    v->item[e].u.vec=allocate_n_array(sp,t);
  }
  return v;
}

struct vector *sum_arrays(
	int num_of_arrays,
	struct svalue *arrays,
	struct svalue *fun)
{
  int d,e,size;
  struct vector *res;
  struct svalue *ret;

  size=arrays->u.vec->size;

  res=allocate_array(size);
  for(e=0;e<size;e++)
  {
    for(d=0;d<num_of_arrays;d++)
    {
      if(arrays[d].u.vec->size>e)
        push_svalue(& arrays[d].u.vec->item[e]);
      else
        push_number(0);
    }
    if(fun->u.ob->flags & O_DESTRUCTED)
    {
      free_vector(res);
      error("Object used by sum_arrays destructed.\n",0,0,0,0,0,0,0,0,0);
    }
    ret=apply_lambda(fun,num_of_arrays);
    if(!ret)
    {
      free_vector(res);
      error("Function used by sum_arrays() not found.\n",0,0,0,0,0,0,0,0,0);
    }
    assign_svalue(& res->item[e],ret);
  }
  return res;
}


struct vector *file_stat(char *file)
{
  struct vector *tmp;
  struct stat    stbuf;  
  char          *p;

  p = file;
  if(!batch_mode)
    while(*p && *p == '/')
      p++;

  if (stat(p, &stbuf) == -1)
    return 0;
  
  tmp = allocate_array(4);
  
  tmp->item[0].u.number = stbuf.st_mode;
  tmp->item[1].u.number = stbuf.st_size;
  tmp->item[2].u.number = stbuf.st_atime;
  tmp->item[3].u.number = stbuf.st_mtime;

  tmp->types=T_NUMBER;
  return tmp;
}
