#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <memory.h>
#include "config.h"
#include "lint.h"
#include "array.h"
#include "interpret.h"
#include "object.h"
#include "comm1.h"
#include "exec.h"
#include "call_out.h"
#include "sent.h"
#include "wiz_list.h"
#include "main.h"
#include "stralloc.h"

/*
 * stralloc.c - string management.
 *
 * All strings are stored in an extensible hash table, with reference counts.
 * free_string decreases the reference count; if it gets to zero, the string
 * will be deallocated.  add_string increases the ref count if it finds a
 * matching string, or allocates it if it cant.  There is no way to allocate
 * a string of a particular size to fill later (hash wont work!), so you'll
 * have to copy things freom a static (or malloced and later freed) buffer -
 * that is, if you want to avoid space leaks...
 *
 * Current overhead:
 *	12 bytes per string (next pointer, and 2 shorts for length and refs)
 *	Strings are nearly all fairly short, so this is a significant overhead-
 *	there is also the 4 byte malloc overhead and the fact that malloc
 *	generally allocates blocks which are a power of 2 (should write my
 *	own best-fit malloc specialised to strings); then again, GNU malloc
 *	is bug free...
 */
#ifdef DEBUG
/* #define ALERT "" */
#endif

extern int d_flag;
#ifdef ALERT
void Alert(char *s)
{
  fprintf(stderr,"Alert: '%s' done.\n",s);
}
void Alert2(char *s)
{
  if(d_flag>6) Alert(s);
}
#endif


/*
 * there is also generic hash table management code, but strings can be shared
 * (that was the point of this code), will be unique in the table,
 * require a reference count, and are malloced, copied and freed at
 * will by the string manager.  Besides, I wrote this code first :-).
 * Look at htable.c for the other code.  It uses the Hash() function
 * defined here, and requires hashed objects to have a pointer to the
 * next element in the chain (which you specify when you call the functions).
 */

#define	MAXSHORT	(1 << (sizeof(int)*8 - 2))
#define	WORD_ALIGN_BIT	0x3	/* these are 0 for aligned ptrs */


struct shared
{
  char *next;
  int refs;
#ifdef DEBUG
  int extra_ref;
#endif
  int size;
  char str[1]; /* Must be last */
};

#define REFERENCE(X,Y) (BASEOF(X,shared,str)Y)

#define SIZE(X) REFERENCE(X,->size)
#define REFS(X) REFERENCE(X,->refs)
#define STR(X) REFERENCE(X,->str)
#define XREF(X) REFERENCE(X,->extra_ref)
#define NEXT(X) REFERENCE(X,->next)


static int num_distinct_strings = 0;
int bytes_distinct_strings = 0;
static int allocd_strings = 0;
static int allocd_bytes = 0;
int overhead_bytes = 0;
static int search_len = 0;
static int num_str_searches = 0;

#ifdef DEBUG
int refchecked=0;
#endif


/*
 * hash table - list of pointers to heads of string chains.
 * Each string in chain has a pointer to the next string and a
 * reference count (char *, int) stored just before the start of the string.
 * HTABLE_SIZE is in config.h, and should be a prime, probably between
 * 1000 and 5000.
 */

char ** base_table = 0;

void init_strings()
{
  int x;
  base_table=(char **)xalloc(sizeof(char *) * HTABLE_SIZE);
  overhead_bytes += (sizeof(char *) * HTABLE_SIZE);

  for (x=0; x<HTABLE_SIZE; x++) base_table[x] = 0;
}

void uninit_strings() { if(base_table) free((char *)base_table); }

/*
 * generic hash function.  This is probably overkill; I haven't checked the
 * stats for different prime numbers, etc.
 */

static int StrHash(char *s,int len)
{
  return hashmem(s, len, 20 , HTABLE_SIZE);
}

/*
 * Looks for a string in the table.  If it finds it, returns a pointer to
 * the start of the string part, and moves the entry for the string to
 * the head of the pointer chain.  One thing (blech!) - puts the previous
 * pointer on the hash chain into fs_prev.
 */

static char * internal_findstring(char *s,int len,int h)
{
  char * curr, *prev;

  curr = base_table[h];
  prev = 0;
  num_str_searches++;

  while (curr)
  {
    search_len++;
    if (len==SIZE(curr) && !memcmp(curr, s,len)) /* found it */
    {
      if (prev) {		/* not at head of list */
	NEXT(prev) = NEXT(curr);
	NEXT(curr) = base_table[h];
	base_table[h] = curr;
      }
#ifdef DEBUG
      if(REFS(curr)<1)
      {
	fatal("String with no references.\n");
      }
#endif
      return(curr);		/* pointer to string */
    }
    prev = curr;
    curr = NEXT(curr);
  }
	
  return(0);			/* not found */
}

char *findstring(char *foo)
{
  int l;
  l=strlen(foo);
  return internal_findstring(foo,l,StrHash(foo,l));
}

/* this one looks for strings that was already shared only */
char * find_shared_string(char *s,int h)
{
  char * curr, *prev;

  curr = base_table[h];
  prev = 0;
  num_str_searches++;

  while (curr)
  {
    search_len++;
    if (curr == s) /* found it */
    {
      if (prev) {		/* not at head of list */
	NEXT(prev) = NEXT(curr);
	NEXT(curr) = base_table[h];
	base_table[h] = curr;
      }
#ifdef DEBUG
      if(REFS(curr)<1)
      {
	fatal("String with no references.\n");
      }
#endif
      return(curr);		/* pointer to string */
    }
    prev = curr;
    curr = NEXT(curr);
  }
	
  return(0);			/* not found */
}

char *debug_findstring(char *foo)
{
  return find_shared_string(foo,StrHash(foo,SIZE(foo)));
}

/*
 * Make a space for a string.  This is rather nasty, as we want to use
 * alloc/free, but malloc overhead is generally severe.  Later, we
 * will have to compact strings...
 */

/* note that begin_shared_string expects the _exact_ size of the string,
 * not the maximum size
 */
char *begin_shared_string(int len)
{
  struct shared *t;
  t=(struct shared *)xalloc(len + sizeof(struct shared));
  t->size=len;
  return t->str;
}

static char * alloc_new_string(char *string,int len,int h)
{
  char *s;
  s=begin_shared_string(len);
  memcpy(s, string,len);
  s[len]=0;
  REFS(s) = 0;
#ifdef DEBUG
  XREF(s)=0;
#endif
  NEXT(s) = base_table[h];
  base_table[h] = s;
  num_distinct_strings++;
  bytes_distinct_strings += 4 + (len +3) & ~3;
  overhead_bytes += sizeof(struct shared)-1;
  return s;
}

char *end_shared_string(char *s)
{
  int len,h;
  char *s2;

  len=SIZE(s);
  h=StrHash(s,len);
  
  if(s2=internal_findstring(s,len,h))
  {
    free((char *)REFERENCE(s,));
    s=s2;
  }else{
    s[len]=0;
    REFS(s) = 0;
#ifdef DEBUG
    XREF(s)=0;
#endif
    NEXT(s) = base_table[h];
    base_table[h] = s;
    num_distinct_strings++;
    bytes_distinct_strings += 4 + (len +3) & ~3;
    overhead_bytes += sizeof(struct shared)-1;
  }
  REFS(s)++;
#ifdef DEBUG
  XREF(s)++;
#endif
  allocd_strings++;
  allocd_bytes += 4 + (len + 3) & ~3;
  return s;
}

char * make_shared_string(char *str)
{
  char * s;
  int len=strlen(str);
  int h=StrHash(str,len);

#ifdef ALERT
  if(!strcmp(str,ALERT)) Alert("make_shared_string");
#endif

  s = internal_findstring(str,len,h);
  if (!s) s = alloc_new_string(str,len,h);
  REFS(s)++;
#ifdef DEBUG
  XREF(s)++;
#endif
  allocd_strings++;
  allocd_bytes += 4 + (len + 3) & ~3;
  return(s);
}

char * make_shared_binary_string(char *str,int len)
{
  char * s;
  int h=StrHash(str,len);

#ifdef ALERT
  if(!strcmp(str,ALERT)) Alert("make_shared_binary_string");
#endif

  s = internal_findstring(str,len,h);
  if (!s) s = alloc_new_string(str,len,h);
  REFS(s)++;
#ifdef DEBUG
  XREF(s)++;
#endif
  allocd_strings++;
  allocd_bytes += 4 + (len + 3) & ~3;
  return(s);
}

int my_strlen(struct svalue *s)
{
#ifdef DEBUG
  if(s->type!=T_STRING)
    fatal("Bad type to internal strlen.\n");
#endif
  if(s->subtype==STRING_SHARED) return SIZE(s->u.string);
  return strlen(s->u.string);
}

int my_string_is_equal(struct svalue *a,struct svalue *b)
{
#ifdef DEBUG
  if(a->type!=T_STRING)
    fatal("Bad type to internal strcmp.\n");
  if(b->type!=T_STRING)
    fatal("Bad type to internal strcmp.\n");
#endif
  if(a->subtype==STRING_SHARED && 
     b->subtype==STRING_SHARED)
    return a->u.string == b->u.string;
  return !strcmp(a->u.string,b->u.string);

}

int my_strcmp(struct svalue *a,struct svalue *b)
{
  int tmp;
#ifdef DEBUG
  if(a->type!=T_STRING)
    fatal("Bad type to internal strcmp.\n");
  if(b->type!=T_STRING)
    fatal("Bad type to internal strcmp.\n");
#endif
  if(a->u.string == b->u.string) return 0;
  if(a->subtype==STRING_SHARED && 
     b->subtype==STRING_SHARED)
  {
    if(SIZE(a->u.string)>SIZE(b->u.string))
    {
      tmp=memcmp(a->u.string,b->u.string,SIZE(b->u.string));
      if(tmp) return tmp;
      return 1;
    }else if(SIZE(a->u.string)<SIZE(b->u.string)){
      tmp=memcmp(a->u.string,b->u.string,SIZE(a->u.string));
      if(tmp) return tmp;
      return -1;
    }else{
      return memcmp(a->u.string,b->u.string,SIZE(a->u.string));
    }
  }
  return strcmp(a->u.string,b->u.string);
}

char * copy_shared_string(char *str)
{
#ifdef ALERT
  if(!strcmp(str,ALERT)) Alert("copy_shared_string");
#endif

#ifdef DEBUG
   if(debug_findstring(str)!=str)
    fatal("Copy shared string on non shared string.\n",0,0,0,0,0,0,0,0,0);
#endif
  REFS(str)++;
#ifdef DEBUG
  XREF(str)++;
#endif
  allocd_strings++;
  allocd_bytes += 4 + (SIZE(str) + 3) & ~3;
  return(str);
}

/*
 * free_string - reduce the ref count on a string.  Various sanity
 * checks applied, the best of them being that a add_string allocated
 * string will point to a word boundary + sizeof(char *)+sizeof(int),
 * since malloc always allocates on a word boundary.
 * On systems where a short is 1/2 a word, this gives an easy check
 * to see whather we did in fact allocate it.
 *
 * Don't worry about the overhead for all those checks!
 */

/*
 * function called on free_string detected errors; things return checked(s).
 */

static void checked(s, str) char * s, *str;
{
  fprintf(stderr, "%s (\"%s\")\n", s, str);
  fatal(s,0,0,0,0,0,0,0,0,0); /* brutal - debugging */
}

void free_string(char *str)
{
  char * s;
  int h=StrHash(str,SIZE(str));

#ifdef ALERT
  if(!strcmp(str,ALERT)) Alert("free_string");
#endif

  allocd_strings--;
  allocd_bytes -= 4 + (SIZE(str) + 3) & ~3;

#ifdef	DEBUG
#ifdef	dcheck			/* GNU malloc range check flag */
{
  int align;
  align = (((int)str) - sizeof(char *) - sizeof(int)) & WORD_B_MASK;
  if (align)
    checked("Free string: improperly aligned string!", str);
}
#endif /* dcheck */
#endif

#ifndef	DEBUG
  if(--REFS(str)) return;
  s = find_shared_string(str,h); /* moves it to head of table if found */
#else
  s = find_shared_string(str,h); /* moves it to head of table if found */
  if (!s) {
    checked("Free string: not found in string table!", str);
    return;
  }
  if (s != str) {
    checked("Free string: string didnt hash to the same spot!", str);
    return;
  }

  if (REFS(s) <= 0) {
    checked("Free String: String refs zero or -ve!", str);
    return;
  }
#ifdef DEBUG
  if(REFS(s) != XREF(s) && refchecked)
  {
    fprintf(stderr,"Freeing string with strange xrefs '%s'!\n",s);
  }
  XREF(s)--;
#endif
  REFS(s)--;
  if (REFS(s) > 0) return;
#endif /* DEBUG */


  /* It will be at the head of the hash chain */
  base_table[h] = NEXT(s);
  num_distinct_strings--;
  /* We know how much overhead malloc has */
  bytes_distinct_strings-= 4 + (SIZE(s) + 3) & ~3;
  overhead_bytes -= sizeof(int) + sizeof(char *);
  free((char *)REFERENCE(s,));

  return;
}

/*
 * you think this looks bad!  and we didn't even tell them about the
 * GNU malloc overhead!  tee hee!
 */

char *add_string_status(int verbose)
{
  char b[200];
  init_buf();

  if (verbose)
  {
    my_strcat("\nShared string hash table:\n");
    my_strcat("-------------------------\t Strings    Bytes\n");
  }
  sprintf(b,"Strings malloced\t\t%8d %8d + %d overhead\n",
		num_distinct_strings, bytes_distinct_strings, overhead_bytes);
  my_strcat(b);
  if (verbose)
  {
    sprintf(b,"Total asked for\t\t\t%8d %8d\n",
		    allocd_strings, allocd_bytes);
    my_strcat(b);
    sprintf(b,"Space actually required/total string bytes %d%%\n",
		    (bytes_distinct_strings + overhead_bytes)*100 / allocd_bytes);
    my_strcat(b);
    sprintf(b,"Searches: %d    Average search length: %6.3f\n",
		    num_str_searches, (double)search_len / num_str_searches);
    my_strcat(b);
  }
  return free_buf();
/*  return(bytes_distinct_strings + overhead_bytes); */
}

/* DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG */

#ifdef DEBUG

int check_svalues(struct svalue *s,int number,int flag,char *st);

int check_string(char *c, int flag,char *s)
{
  char *t;
  if(!c) return 0;
#ifdef ALERT
  if(!strcmp(c,ALERT))
  {
    char buf[200];
    sprintf(buf,"check_string (flag=%d)",flag);
    Alert2(buf);
  }
#endif
  if(flag==2)
  {
    XREF(c)++;
  }else{
    t=debug_findstring(c);
   if(!t) fatal("Shared string not shared.\n",0,0,0,0,0,0,0,0,0);
    if(t!=c)
      fatal("Shared string not same as one in hashtable.\n",0,0,0,0,0,0,0,0,0);
    XREF(c)=0;
  }
  if(s && s==c)
  {
    fprintf(stderr,"Found '%s'\n",s);
    return 1;
  }
  return 0;
}

char *fulpatch1(struct object *o);
char *fulpatch2(struct object *o);
struct sentence *fulpatch3(struct object *o);

int check_program(struct program *p,int flag,char *s)
{
  int e,ret;
  struct vector *v;

#ifdef ALERT
  if(!strcmp(ALERT,p->name))
  {
    Alert2("foo on bar");
  }
#endif
  if(!p || p->extra_ref==-flag) return 0;
  ret=0;
  if(check_string(p->name,flag,s))
  {
    fprintf(stderr,"Was the name of a program.\n");
    ret=1;
  }
  p->extra_ref=-flag;
  for(e=0;e<p->num_variables;e++)
  {
    if(check_string(p->variable_names[e].name,flag,s))
    {
      fprintf(stderr,"It's the name of a variable in %s.\n",p->name);
      ret=1;
    }
  }
  for(e=0;e<p->num_functions;e++)
  {
    if(check_string(p->functions[e].name,flag,s))
    {
      fprintf(stderr,"It's the name of a function in %s.\n",p->name);
      ret=1;
    }
  }
  for(e=0;e<p->num_strings;e++)
  {
    if(check_string(p->strings[e],flag,s))
    {
      fprintf(stderr,"It's a string in %s.\n",p->name);
      ret=1;
    }
  }
  for(e=0;e<p->num_inherited;e++)
  {
    if(check_program(p->inherit[e].prog,flag,s))
    {
      fprintf(stderr,"Was inherited by '%s'.\n",p->name);
      ret=1;
    }
  }
  for(e=0;e<p->num_switch_mappings;e++)
  {
    v=p->switch_mappings[e];
    if(check_svalues(v->item,v->size,flag,s))
    {
      fprintf(stderr,"Was in a switch mapping of %s.\n",p->name);
      ret=1;
    }
  }

  if(check_svalues(p->constants,p->num_constants,flag,s))
  {
    fprintf(stderr,"Was in a program constant of %s.\n",p->name);
    ret=1;
  }
  return ret;
}

int check_object(struct object *ob, int flag,char *s)
{
  int e,ret;

  if(ob->extra_ref==-flag) return 0;
  ob->extra_ref=-flag;
  ret=0;
  if(check_svalues(ob->variables, ob->prog->num_variables,flag,s))
  {
    fprintf(stderr,"Found in variable\n");
    ret=1;
  }

  if(check_string(ob->obj_index,flag,s))
  {
    fprintf(stderr,"It's the index of an object.\n");
    ret=1;
  }
  if(ob->sentences)
  {
    for(e=0;e<ob->sentences->slots;e++)
    {
      if(!ob->sentences->sent[e].match) continue;
      if(check_svalues(&(ob->sentences->sent[e].function),1,flag,s))
      {
	fprintf(stderr,"It's the function name in an add_action\n");
	ret=1;
      }
    }
  }
  if(check_program(ob->prog,flag,s))
  {
    fprintf(stderr,"Found in program %s\n",ob->prog->name);
    ret=1;
  }
  return ret;
}

int check_array(struct vector *v,int flag,char *s)
{
  if(v->extra_ref==-flag) return 0;
  v->extra_ref=-flag;
  if(flag==2 && v->user) v->user->debug_indices+=v->size;
  return check_svalues(v->item,v->size,flag,s);
}

int check_svalues(struct svalue *s,int number,int flag,char *st)
{
  int e,i;
  i=0;
  for(e=0;e<number;e++,s++)
  {
    switch(s->type)
    {
    case T_STRING:
      switch(s->subtype)
      {
      case STRING_SHARED:
	if(check_string(s->u.string,flag,st))
	{
	  fprintf(stderr,"In shared string svalue.\n");
	  i=1;
	}
	break;

      case STRING_TEMPORARY:
	fatal("Temporary string found in backend loop.\n");
      }
      break;

    case T_MAPPING:
    case T_LIST:
    case T_POINTER:
    case T_ALIST_PART:
      if(check_array(s->u.vec,flag,st))
      {
	switch(s->type)
	{
	case T_MAPPING: fprintf(stderr,"In mapping.\n"); break;
	case T_LIST:  fprintf(stderr,"In list.\n"); break;
	case T_POINTER: fprintf(stderr,"In array.\n"); break;
	case T_ALIST_PART: fprintf(stderr,"In alist part.\n"); break;
	}
	i=1;
      }
      break;

    case T_OBJECT:
      if(check_object(s->u.ob,flag,st))
      {
	fprintf(stderr,"Found in object %s\n",s->u.ob->prog->name);
	i=1;
      }
    }
  }
  return i;
}

int check_str_ref_from_call_outs(int flag,char *s)
{
  int e,ret;
  for(ret=e=0;e<num_pending_calls;e++)
  {
    if(check_object(pending_calls[e]->command_giver,flag,s))
    {
      fprintf(stderr,"Was command giver in call_out.\n");
      ret=1;
    }
    if(check_array(pending_calls[e]->args,flag,s))
    {
      fprintf(stderr,"Was in argument list for call_out.\n");
      ret=1;
    }
  }
  return ret;
}

extern struct svalue start_of_stack[EVALUATOR_STACK_SIZE];
extern struct svalue *sp;
extern struct svalue stmp;
extern struct object *obj_list;
extern char **inc_list;
extern int inc_list_size;
extern char *simul_efun_file_name;
extern struct svalue ret_value,temp_sval;
extern struct program ** obj_table;
extern struct keyword predefs[],reswords[];
extern int number_predefs,number_reswords;
extern struct wiz_list *all_wiz;
extern char *buffered_string;

void check_even_more_ref_counts(int flag,char *s)
{
  int e,q;
  char *c;
  struct object *ob;
  struct program *p;
  struct wiz_list *w;

  if(d_flag<3 || !base_table) return ;
  refchecked=1;
  check_str_ref_from_call_outs(flag,s);
  q= sp - start_of_stack + 1;
  if(flag==1)
  {
    for(w=all_wiz;w;w=w->next)
    {
      w->debug_indices=0;
    }
  }
  if(check_svalues(start_of_stack,q,flag,s))
    fprintf(stderr,"Was found on stack\n");

  if(check_string(buffered_string,flag,s))
    fprintf(stderr,"Was last added string.\n");

  if(check_svalues(start_of_stack+q,EVALUATOR_STACK_SIZE-q,flag,s))
    fatal("Found string after stack end.\n",0,0,0,0,0,0,0,0,0);

#ifdef YYOPT
  for(e=0;e<number_predefs;e++)
  {
    if(check_string(predefs[e].word,flag,s))
      fprintf(stderr,"Was efun name\n");
  }

  for(e=0;e<number_reswords;e++)
  {
    if(check_string(reswords[e].word,flag,s))
      fprintf(stderr,"Was a reserved word\n");
  }
#endif
  for(e=0;e<OTABLE_SIZE;e++)
  {
    for(p=obj_table[e];p;p=p->next_hash)
    {
      if(check_program(p,flag,s))
      {
	fprintf(stderr,"Found in prog %s\n",p->name);
      }
    }
  }

  for(ob=obj_list; ob; ob = ob->next_all)
  {
    if(check_object(ob,flag,s))
      fprintf(stderr,"Found in object %s\n",ob->prog->name);
  }
  for(e=0;e<inc_list_size;e++)
  {
    if(check_string(inc_list[e],flag,s))
      fprintf(stderr,"Found in include path.\n");
  }

  if(check_string(simul_efun_file_name,flag,s))
    fprintf(stderr,"Is the name of the simul_efun file.\n");

  if(check_svalues(&ret_value,1,flag,s))
    fprintf(stderr,"Was last returned value.\n");
    
  if(check_svalues(&temp_sval,1,flag,s))
    fprintf(stderr,"Was list-temporary-svalue.\n");
    
  if(check_svalues(&stmp,1,flag,s))
    fprintf(stderr,"Was in insert_alist quickfix (stmp).\n");

  if(flag==1)
  {
    check_even_more_ref_counts(flag+1,0);
    return;
  }
  if(flag==1)
  {
    for(w=all_wiz;w;w=w->next)
    {
      if(w->debug_indices!=(w->size_array))
      {
	fprintf(stderr,"Warning index in arrays not right for %s (%d!=%d)\n",
	       w->name,w->size_array,w->debug_indices);
      }
    }
  }
  if(flag>2) return;
  
  for(e=0;e<HTABLE_SIZE;e++)
  {
    c=base_table[e];
    while(c)
    {
      if(XREF(c)!=REFS(c))
      {
	flag++;
        fprintf(stderr,"Warning: '%s' had %d refs, expected %d\n",c,REFS(c),XREF(c));
	check_even_more_ref_counts(flag,c);
	if(d_flag>4) fatal("Error in string refs.\n",0,0,0,0,0,0,0,0,0);
      }
      c=NEXT(c);
    }
  }
}
#endif
